#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov 22 09:41:26 2021

@author: santi

This is an interactive, Mayavi based version of the Jupyter Notebook on Solar 
System Simulations using NASA Horizons data. Using the Sun, the main 8 planets and 
their current locations, get initial data on and simulate the rough trajectory 
of any object such as a comet, asteroid, dwarf planet etc. in NASA Horizons.

"""

import numpy as np
import scipy as sp
import time
import scipy.constants as cs
from astropy.time import Time
from astroquery.jplhorizons import Horizons

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from traits.api import HasTraits, Instance, Property, Enum
from traitsui.api import View, Item, HSplit, VSplit, InstanceEditor
from tvtk.pyface.scene_editor import SceneEditor
from mayavi.core.ui.engine_view import EngineView
from mayavi.tools.mlab_scene_model import MlabSceneModel

#Conversion Units
AU = 149597870700
D = 24*60*60

#define functions for the program

#Vectorial acceleration function
def a_t( r, m, epsilon):
    """
    Function of matrices returning the gravitational acceleration
    between N-bodies within a system with positions r and masses m
    -------------------------------------------------
    r  is a N x 3 matrix of object positions
    m is a N x 1 vector of object masses
    epsilon is the softening factor to prevent numerical errors
    a is a N x 3 matrix of accelerations
    -------------------------------------------------
    """  
    G = cs.gravitational_constant/(AU**3)
    # positions r = [x,y,z] for all bodies in the N-Body System
    x = r[:,0:1]
    y = r[:,1:2]
    z = r[:,2:3]
    # matrices that store each pairwise body separation for each [x,y,z] direction: r_j - r_i
    dx = x.T - x
    dy = y.T - y
    dz = z.T - z
    #matrix 1/r^3 for the absolute value of all pairwise body separations together and 
    #resulting acceleration components in each [x,y,z] direction  
    inv_r3 = (dx**2 + dy**2 + dz**2 + epsilon**2)**(-1.5)
    ax = G * (dx * inv_r3) @ m
    ay = G * (dy * inv_r3) @ m
    az = G * (dz * inv_r3) @ m
    # pack together the three acceleration components
    a = np.hstack((ax,ay,az))
    return a

def omega(N, t_max, dt):
    """
    Dummy acceleration function to give an estimate of the total memory
    consumption for a simulation with N bodies, total simulation time 
    t_max and timestep dt
    -------------------------------------------------
    N is the amount of bodies in the simulation
    t_max is the total simulation time in years
    dt is the timestep for each integration in days
    -------------------------------------------------
    """
    epsilon = 1
    G = 1
    r = np.ones((N,3))
    v = np.ones((N,3)) 
    """although the velocities aren't actually taken into account for computing the 
    acceleration, they will be stored as Nx3 matrices for exactly as many iterations 
    in the integration loop itself and take up memory accordingly"""
    m = np.ones((N,1))
    # positions r = [x,y,z] for all bodies in the N-Body System
    x = r[:,0:1]
    y = r[:,1:2]
    z = r[:,2:3]
    # matrices that store each pairwise body separation for each [x,y,z] direction: r_j - r_i
    dx = x.T - x
    dy = y.T - y
    dz = z.T - z
    #matrix 1/r^3 for the absolute value of all pairwise body separations together and 
    #resulting acceleration components in each [x,y,z] direction 
    inv_r3 = (dx**2 + dy**2 + dz**2 + epsilon**2)**(-1.5)
    ax = G * (dx * inv_r3) @ m
    ay = G * (dy * inv_r3) @ m
    az = G * (dz * inv_r3) @ m
    # pack together the three acceleration components
    a = np.hstack((ax,ay,az))
    # sum the memory usage of each matrix storing the positions, distances and accelerations
    memory_usage_per_iteration = r.nbytes + v.nbytes + x.nbytes + y.nbytes + z.nbytes + dx.nbytes + dy.nbytes + dz.nbytes + ax.nbytes + ay.nbytes + az.nbytes + inv_r3.nbytes + a.nbytes 
    total_memory_usage = memory_usage_per_iteration * (t_max)/(dt*10e6) 
    return total_memory_usage #in megabytes




def simulate_solar_system(N,dN,starting_values): #
    t0_sim_start = time.time()
    t = 0
    t_max = 365*24*60*60*N #N year simulation time
    dt = 60*60*24*dN #dN day time step
    epsilon_s = 0.01 #softening default value
    r_i = starting_values[0]/AU
    v_i = starting_values[1]/AU
    m_i = starting_values[2]
    a_i = a_t(r_i, m_i, epsilon_s)
    ram_usage_estimate = omega(len(r_i), t_max, dt) #returns the estimated ram usage for the simulation
    # Simulation Main Loop using a Leapfrog Kick-Drift-Kick Algorithm
    k = int(t_max/dt)
    r_save = np.zeros((r_i.shape[0],3,k+1))
    r_save[:,:,0] = r_i
    for i in range(k):
        # (1/2) kick
        v_i += a_i * dt/2.0
        # drift
        r_i += v_i * dt
        # update accelerations
        a_i = a_t(r_i, m_i, epsilon_s)
        # (2/2) kick
        v_i += a_i * dt/2.0
        # update time
        t += dt
        #update list
        r_save[:,:,i+1] = r_i
    sim_time = time.time()-t0_sim_start
    print('The required computation time for the N-Body Simulation was', round(sim_time,3), 'seconds. The estimated memory usage was', round(ram_usage_estimate,3), 'megabytes of RAM.')
    return r_save

#Main Program Loop

from traits.api import *
from traitsui.api import *
from mayavi import mlab
import random


class SolarSystemObjects( HasTraits ):
   """Scene Creation"""
   # The scene model.
   scene = Instance(MlabSceneModel, ())

   # The mayavi engine view.
   engine_view = Instance(EngineView)



   ######################
   view = View(HSplit(VSplit(Item(name='engine_view',
                                   style='custom',
                                   resizable=True,
                                   show_label=False
                                   ),
                               Item(name='scene',
                                    editor=SceneEditor(),
                                    show_label=False,
                                    resizable=True,
                                    height=500,
                                    width=500),
                        ),
                resizable=True,
                scrollable=True
                ))
    
   """ Simulation object """
   PlanetaryObjects = Enum('Inner Planets', 'Outer Planets', 'Inner and outer Planets',
      desc="Choose which planets to account for in the simulation",
      label="Planets", )
   t_max = CInt(1,
      desc="The maximal simulation time since the start, in years",
      label="Simulation Time (years)", )
   dt = CInt(1,
      desc="The timestep for the simulation, in days",
      label="Timestep (days)", )
   t_0 = Str('XXXX-XX-XX',
      desc="Simulation starting date in Y/M/D",
      label="Simulation Start (Date in Y/M/D)", )
   obj_id = Str('',
                desc="The Horizons ID of the object to simulate",
                label="HorizonsID")
   
   update_preview = Button("Update Preview", show_label=False)
   
   def simulate(self):
       "Simulates the solar system in the specified region with the added object in Horizons"
       r_list = []
       v_list = []
       m_list = [[1.989e30],[3.285e23],[4.867e24],[5.972e24],[6.39e23],[1.8989e27],[5.683e26],[8.681e25],[1.024e26],[1.309e22]] #Object masses for Sun-Pluto
       plot_colors = ['green','brown','orange','blue','red','red','orange','cyan','blue','brown']
       plot_labels = ['Barycenter','Mercury','Venus','Earth','Mars','Jupiter','Saturn','Uranus','Neptune','Pluto']
       if self.PlanetaryObjects == str("Inner Planets"):
           m_list = m_list[0:5]
           for i in range(0,5):
               obj = Horizons(id=i, location="@sun", epochs=Time(self.t_0).jd, id_type='id').vectors()
               r_obj = [obj['x'][0], obj['y'][0], obj['z'][0]]
               v_obj = [obj['vx'][0], obj['vy'][0], obj['vz'][0]]
               r_list.append(r_obj)
               v_list.append(v_obj)
           
       elif self.PlanetaryObjects == str("Outer Planets"):
           m_outer = m_list[5:10]
           m_list = m_list[0]
           m_list.append(m_outer)
           obj = Horizons(id=0, location="@sun", epochs=Time(self.t_0).jd, id_type='id').vectors()
           r_obj = [obj['x'][0], obj['y'][0], obj['z'][0]]
           v_obj = [obj['vx'][0], obj['vy'][0], obj['vz'][0]]
           r_list.append(r_obj)
           v_list.append(v_obj)
           for i in range(5,10):
               obj = Horizons(id=i, location="@sun", epochs=Time(self.t_0).jd, id_type='id').vectors()
               r_obj = [obj['x'][0], obj['y'][0], obj['z'][0]]
               v_obj = [obj['vx'][0], obj['vy'][0], obj['vz'][0]]
               r_list.append(r_obj)
               v_list.append(v_obj)
       elif self.PlanetaryObjects == str("Inner and outer Planets"):
           for i in range(0,10):
               obj = Horizons(id=i, location="@sun", epochs=Time(self.t_0).jd, id_type='id').vectors()
               r_obj = [obj['x'][0], obj['y'][0], obj['z'][0]]
               v_obj = [obj['vx'][0], obj['vy'][0], obj['vz'][0]]
               r_list.append(r_obj)
               v_list.append(v_obj)
       def add_simulation_object(Id_obj,t_0,m_obj, plot_color, plot_label):
           obj = Horizons(id=Id_obj, location="@sun", epochs=Time(t_0).jd, id_type='id').vectors()
           r_obj = [obj['x'][0], obj['y'][0], obj['z'][0]]
           v_obj = [obj['vx'][0], obj['vy'][0], obj['vz'][0]]
           r_list.append(r_obj)
           v_list.append(v_obj)
           m_list.append([m_obj])
           plot_colors.append(plot_color)
           plot_labels.append(plot_label)
       try:
           add_simulation_object(self.obj_id,self.t_0,1e12, 'black',self.obj_id)
       except ValueError:
           print("The ID is not valid.  Try again...")
       #Convert object staring value lists to numpy
       r_i = np.array(r_list)*AU
       v_i = np.array(v_list)*AU/D
       m_i = np.array(m_list)
       #pack together as list for the simulation function
       horizons_data = [r_i,v_i,m_i]
       #Run simulation for t_max years at a dt day time-step
       r_save = simulate_solar_system(self.t_max,self.dt,horizons_data)
       return r_save
    
       def __init__(self, **traits):
           HasTraits.__init__(self, **traits)
           self.engine_view = EngineView(engine=self.scene.engine)

           # Hook up the current_selection to change when the one in the engine
           # changes.  This is probably unnecessary in Traits3 since you can show
           # the UI of a sub-object in T3.
           self.scene.engine.on_trait_change(self._selection_change,
                                          'current_selection')

           r_save = self.simulate()
    

    #      Plot of Solar System Barycenter
           mlab.figure(bgcolor=(1.0,1.0,1.0),fgcolor=(0.0,0.0,0.0))
           for i in range(0,len(r_i)-1): #Plots any additional objects
               mlab.plot3d(r_save[i,0,:],r_save[i,1,:],r_save[i,2,:], line_width=0.5, color=(random.uniform(0, 1), random.uniform(0, 1), random.uniform(0, 1)))
           mlab.plot3d(r_save[len(r_i)-1,0,:],r_save[len(r_i)-1,1,:],r_save[len(r_i)-1,2,:], line_width=0.5, color=(0.0, 0.0, 0.0))
           mlab.axes(xlabel='AU', ylabel='AU', zlabel='AU',nb_labels=5)
           mlab.show()
      
    #---------------------------------------------------------------------------
    # The layout of the dialog created
    #---------------------------------------------------------------------------


if  __name__ == "__main__":
   solar_sim_objects = SolarSystemObjects()
   solar_sim_objects.configure_traits()